#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from abc import ABC, abstractmethod

from eland import SortOrder
from eland.actions import HeadAction, TailAction, SortIndexAction
from eland.arithmetics import ArithmeticSeries


# -------------------------------------------------------------------------------------------------------------------- #
# Tasks                                                                                                                #
# -------------------------------------------------------------------------------------------------------------------- #
class Task(ABC):
    """
    Abstract class for tasks

    Parameters
    ----------
        task_type: str
            The task type (e.g. head, tail etc.)
    """

    def __init__(self, task_type):
        self._task_type = task_type

    @property
    def type(self):
        return self._task_type

    @abstractmethod
    def resolve_task(self, query_params, post_processing, query_compiler):
        pass

    @abstractmethod
    def __repr__(self):
        pass


class SizeTask(Task):
    def __init__(self, task_type):
        super().__init__(task_type)

    @abstractmethod
    def size(self):
        # must override
        pass


class HeadTask(SizeTask):
    def __init__(self, sort_field, count):
        super().__init__("head")

        # Add a task that is an ascending sort with size=count
        self._sort_field = sort_field
        self._count = count

    def __repr__(self):
        return "('{}': ('sort_field': '{}', 'count': {}))".format(self._task_type, self._sort_field, self._count)

    def resolve_task(self, query_params, post_processing, query_compiler):
        # head - sort asc, size n
        # |12345-------------|
        query_sort_field = self._sort_field
        query_sort_order = SortOrder.ASC
        query_size = self._count

        # If we are already postprocessing the query results, we just get 'head' of these
        # (note, currently we just append another head, we don't optimise by
        # overwriting previous head)
        if len(post_processing) > 0:
            post_processing.append(HeadAction(self._count))
            return query_params, post_processing

        if query_params['query_sort_field'] is None:
            query_params['query_sort_field'] = query_sort_field
        # if it is already sorted we use existing field

        if query_params['query_sort_order'] is None:
            query_params['query_sort_order'] = query_sort_order
        # if it is already sorted we get head of existing order

        if query_params['query_size'] is None:
            query_params['query_size'] = query_size
        else:
            # truncate if head is smaller
            if query_size < query_params['query_size']:
                query_params['query_size'] = query_size

        return query_params, post_processing

    def size(self):
        return self._count


class TailTask(SizeTask):
    def __init__(self, sort_field, count):
        super().__init__("tail")

        # Add a task that is descending sort with size=count
        self._sort_field = sort_field
        self._count = count

    def __repr__(self):
        return "('{}': ('sort_field': '{}', 'count': {}))".format(self._task_type, self._sort_field, self._count)

    def resolve_task(self, query_params, post_processing, query_compiler):
        # tail - sort desc, size n, post-process sort asc
        # |-------------12345|
        query_sort_field = self._sort_field
        query_sort_order = SortOrder.DESC
        query_size = self._count

        # If this is a tail of a tail adjust settings and return
        if query_params['query_size'] is not None and \
                query_params['query_sort_order'] == query_sort_order and \
                post_processing == ['sort_index']:
            if query_size < query_params['query_size']:
                query_params['query_size'] = query_size
            return query_params, post_processing

        # If we are already postprocessing the query results, just get 'tail' of these
        # (note, currently we just append another tail, we don't optimise by
        # overwriting previous tail)
        if len(post_processing) > 0:
            post_processing.append(TailAction(self._count))
            return query_params, post_processing

        # If results are already constrained, just get 'tail' of these
        # (note, currently we just append another tail, we don't optimise by
        # overwriting previous tail)
        if query_params['query_size'] is not None:
            post_processing.append(TailAction(self._count))
            return query_params, post_processing
        else:
            query_params['query_size'] = query_size
        if query_params['query_sort_field'] is None:
            query_params['query_sort_field'] = query_sort_field
        if query_params['query_sort_order'] is None:
            query_params['query_sort_order'] = query_sort_order
        else:
            # reverse sort order
            query_params['query_sort_order'] = SortOrder.reverse(query_sort_order)

        post_processing.append(SortIndexAction())

        return query_params, post_processing

    def size(self):
        return self._count


class QueryIdsTask(Task):
    def __init__(self, must, ids):
        """
        Parameters
        ----------
        must: bool
            Include or exclude these ids (must/must_not)

        ids: list
            ids for the filter
        """
        super().__init__("query_ids")

        self._must = must
        self._ids = ids

    def resolve_task(self, query_params, post_processing, query_compiler):
        query_params['query'].ids(self._ids, must=self._must)

        return query_params, post_processing

    def __repr__(self):
        return "('{}': ('must': {}, 'ids': {}))".format(self._task_type, self._must, self._ids)


class QueryTermsTask(Task):
    def __init__(self, must, field, terms):
        """
        Parameters
        ----------
        must: bool
            Include or exclude these ids (must/must_not)

        field: str
            field_name to filter

        terms: list
            field_values for filter
        """
        super().__init__("query_terms")

        self._must = must
        self._field = field
        self._terms = terms

    def __repr__(self):
        return "('{}': ('must': {}, 'field': '{}', 'terms': {}))".format(self._task_type, self._must, self._field,
                                                                         self._terms)

    def resolve_task(self, query_params, post_processing, query_compiler):
        query_params['query'].terms(self._field, self._terms, must=self._must)

        return query_params, post_processing


class BooleanFilterTask(Task):
    def __init__(self, boolean_filter):
        """
        Parameters
        ----------
        boolean_filter: BooleanFilter or str
            The filter to apply
        """
        super().__init__("boolean_filter")

        self._boolean_filter = boolean_filter

    def __repr__(self):
        return "('{}': ('boolean_filter': {}))".format(self._task_type, repr(self._boolean_filter))

    def resolve_task(self, query_params, post_processing, query_compiler):
        query_params['query'].update_boolean_filter(self._boolean_filter)

        return query_params, post_processing


class ArithmeticOpFieldsTask(Task):
    def __init__(self, display_name, arithmetic_series):
        super().__init__("arithmetic_op_fields")

        self._display_name = display_name

        if not isinstance(arithmetic_series, ArithmeticSeries):
            raise TypeError("Expecting ArithmeticSeries got {}".format(type(arithmetic_series)))
        self._arithmetic_series = arithmetic_series

    def __repr__(self):
        return "('{}': (" \
               "'display_name': {}, " \
               "'arithmetic_object': {}" \
               "))" \
            .format(self._task_type, self._display_name, self._arithmetic_series)

    def update(self, display_name, arithmetic_series):
        self._display_name = display_name
        self._arithmetic_series = arithmetic_series

    def resolve_task(self, query_params, post_processing, query_compiler):
        # https://www.elastic.co/guide/en/elasticsearch/painless/current/painless-api-reference-shared-java-lang.html#painless-api-reference-shared-Math
        """
        "script_fields": {
            "field_name": {
            "script": {
                "source": "doc[self._left_field].value / self._right_field"
            }
            }
        }
        """
        if query_params['query_script_fields'] is None:
            query_params['query_script_fields'] = dict()

        if self._display_name in query_params['query_script_fields']:
            raise NotImplementedError(
                "TODO code path - combine multiple ops '{}'\n{}\n{}\n{}".format(self,
                                                                                query_params['query_script_fields'],
                                                                                self._display_name,
                                                                                self._arithmetic_series.resolve()))

        query_params['query_script_fields'][self._display_name] = {
            'script': {
                'source': self._arithmetic_series.resolve()
            }
        }

        return query_params, post_processing



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.util.testing import assert_series_equal

from eland.tests.common import TestData


class TestDataFrameMetrics(TestData):
    funcs = ['max', 'min', 'mean', 'sum']

    def test_flights_metrics(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        for func in self.funcs:
            pd_metric = getattr(pd_flights, func)(numeric_only=True)
            ed_metric = getattr(ed_flights, func)(numeric_only=True)

            assert_series_equal(pd_metric, ed_metric)

    def test_ecommerce_selected_non_numeric_source_fields(self):
        # None of these are numeric
        columns = ['category', 'currency', 'customer_birth_date', 'customer_first_name', 'user']

        pd_ecommerce = self.pd_ecommerce()[columns]
        ed_ecommerce = self.ed_ecommerce()[columns]

        for func in self.funcs:
            assert_series_equal(getattr(pd_ecommerce, func)(numeric_only=True),
                                getattr(ed_ecommerce, func)(numeric_only=True),
                                check_less_precise=True)

    def test_ecommerce_selected_mixed_numeric_source_fields(self):
        # Some of these are numeric
        columns = ['category', 'currency', 'taxless_total_price', 'customer_birth_date',
                   'total_quantity', 'customer_first_name', 'user']

        pd_ecommerce = self.pd_ecommerce()[columns]
        ed_ecommerce = self.ed_ecommerce()[columns]

        for func in self.funcs:
            assert_series_equal(getattr(pd_ecommerce, func)(numeric_only=True),
                                getattr(ed_ecommerce, func)(numeric_only=True),
                                check_less_precise=True)

    def test_ecommerce_selected_all_numeric_source_fields(self):
        # All of these are numeric
        columns = ['total_quantity', 'taxful_total_price', 'taxless_total_price']

        pd_ecommerce = self.pd_ecommerce()[columns]
        ed_ecommerce = self.ed_ecommerce()[columns]

        for func in self.funcs:
            assert_series_equal(getattr(pd_ecommerce, func)(numeric_only=True),
                                getattr(ed_ecommerce, func)(numeric_only=True),
                                check_less_precise=True)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.testing import assert_index_equal

from eland.tests.common import TestData


class TestDataFrameKeys(TestData):

    def test_ecommerce_keys(self):
        pd_ecommerce = self.pd_ecommerce()
        ed_ecommerce = self.ed_ecommerce()

        pd_keys = pd_ecommerce.keys()
        ed_keys = ed_ecommerce.keys()

        assert_index_equal(pd_keys, ed_keys)

    def test_flights_keys(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        pd_keys = pd_flights.keys()
        ed_keys = ed_flights.keys()

        assert_index_equal(pd_keys, ed_keys)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from abc import ABC, abstractmethod
from io import StringIO

import numpy as np


class ArithmeticObject(ABC):
    @property
    @abstractmethod
    def value(self):
        pass

    @abstractmethod
    def dtype(self):
        pass

    @abstractmethod
    def resolve(self):
        pass

    @abstractmethod
    def __repr__(self):
        pass


class ArithmeticString(ArithmeticObject):
    def __init__(self, value):
        self._value = value

    def resolve(self):
        return self.value

    @property
    def dtype(self):
        return np.dtype(object)

    @property
    def value(self):
        return "'{}'".format(self._value)

    def __repr__(self):
        return self.value


class ArithmeticNumber(ArithmeticObject):
    def __init__(self, value, dtype):
        self._value = value
        self._dtype = dtype

    def resolve(self):
        return self.value

    @property
    def value(self):
        return "{}".format(self._value)

    @property
    def dtype(self):
        return self._dtype

    def __repr__(self):
        return self.value


class ArithmeticSeries(ArithmeticObject):
    def __init__(self, query_compiler, display_name, dtype):
        task = query_compiler.get_arithmetic_op_fields()
        if task is not None:
            self._value = task._arithmetic_series.value
            self._tasks = task._arithmetic_series._tasks.copy()
            self._dtype = dtype
        else:
            aggregatable_field_name = query_compiler.display_name_to_aggregatable_name(display_name)
            if aggregatable_field_name is None:
                raise ValueError(
                    "Can not perform arithmetic operations on non aggregatable fields"
                    "{} is not aggregatable.".format(display_name)
                )

            self._value = "doc['{}'].value".format(aggregatable_field_name)
            self._tasks = []
            self._dtype = dtype

    @property
    def value(self):
        return self._value

    @property
    def dtype(self):
        return self._dtype

    def __repr__(self):
        buf = StringIO()
        buf.write("Series: {} ".format(self.value))
        buf.write("Tasks: ")
        for task in self._tasks:
            buf.write("{} ".format(repr(task)))
        return buf.getvalue()

    def resolve(self):
        value = self._value

        for task in self._tasks:
            if task.op_name == '__add__':
                value = "({} + {})".format(value, task.object.resolve())
            elif task.op_name == '__truediv__':
                value = "({} / {})".format(value, task.object.resolve())
            elif task.op_name == '__floordiv__':
                value = "Math.floor({} / {})".format(value, task.object.resolve())
            elif task.op_name == '__mod__':
                value = "({} % {})".format(value, task.object.resolve())
            elif task.op_name == '__mul__':
                value = "({} * {})".format(value, task.object.resolve())
            elif task.op_name == '__pow__':
                value = "Math.pow({}, {})".format(value, task.object.resolve())
            elif task.op_name == '__sub__':
                value = "({} - {})".format(value, task.object.resolve())
            elif task.op_name == '__radd__':
                value = "({} + {})".format(task.object.resolve(), value)
            elif task.op_name == '__rtruediv__':
                value = "({} / {})".format(task.object.resolve(), value)
            elif task.op_name == '__rfloordiv__':
                value = "Math.floor({} / {})".format(task.object.resolve(), value)
            elif task.op_name == '__rmod__':
                value = "({} % {})".format(task.object.resolve(), value)
            elif task.op_name == '__rmul__':
                value = "({} * {})".format(task.object.resolve(), value)
            elif task.op_name == '__rpow__':
                value = "Math.pow({}, {})".format(task.object.resolve(), value)
            elif task.op_name == '__rsub__':
                value = "({} - {})".format(task.object.resolve(), value)

        return value

    def arithmetic_operation(self, op_name, right):
        # check if operation is supported (raises on unsupported)
        self.check_is_supported(op_name, right)

        task = ArithmeticTask(op_name, right)
        self._tasks.append(task)
        return self

    def check_is_supported(self, op_name, right):
        # supported set is
        # series.number op_name number (all ops)
        # series.string op_name string (only add)
        # series.string op_name int (only mul)
        # series.string op_name float (none)
        # series.int op_name string (none)
        # series.float op_name string (none)

        # see end of https://pandas.pydata.org/pandas-docs/stable/getting_started/basics.html?highlight=dtype for
        # dtype heirarchy
        if np.issubdtype(self.dtype, np.number) and np.issubdtype(right.dtype, np.number):
            # series.number op_name number (all ops)
            return True
        elif np.issubdtype(self.dtype, np.object_) and np.issubdtype(right.dtype, np.object_):
            # series.string op_name string (only add)
            if op_name == '__add__' or op_name == '__radd__':
                return True
        elif np.issubdtype(self.dtype, np.object_) and np.issubdtype(right.dtype, np.integer):
            # series.string op_name int (only mul)
            if op_name == '__mul__':
                return True

        raise TypeError(
            "Arithmetic operation on incompatible types {} {} {}".format(self.dtype, op_name, right.dtype))


class ArithmeticTask:
    def __init__(self, op_name, object):
        self._op_name = op_name

        if not isinstance(object, ArithmeticObject):
            raise TypeError("Task requires ArithmeticObject not {}".format(type(object)))
        self._object = object

    def __repr__(self):
        buf = StringIO()
        buf.write("op_name: {} ".format(self.op_name))
        buf.write("object: {} ".format(repr(self.object)))
        return buf.getvalue()

    @property
    def op_name(self):
        return self._op_name

    @property
    def object(self):
        return self._object



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

__title__ = 'eland'
__description__ = 'Python elasticsearch client to analyse, explore and manipulate data that resides in elasticsearch.'
__url__ = 'https://github.com/elastic/eland'
__version__ = '7.5.1a4'
__maintainer__ = 'Elasticsearch B.V.'
__maintainer_email__ = 'steve.dodson@elastic.co'



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from __future__ import absolute_import

from eland._version import *
from eland.common import *
from eland.client import *
from eland.filter import *
from eland.index import *
from eland.field_mappings import *
from eland.query import *
from eland.operations import *
from eland.query_compiler import *
from eland.ndframe import *
from eland.series import *
from eland.dataframe import *
from eland.utils import *



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import os

import pandas as pd
from pandas.util.testing import (assert_frame_equal, assert_series_equal)

import eland as ed

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# Create pandas and eland data frames
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_DF_FILE_NAME, FLIGHTS_INDEX_NAME, \
    FLIGHTS_SMALL_INDEX_NAME, \
    ECOMMERCE_DF_FILE_NAME, ECOMMERCE_INDEX_NAME

_pd_flights = pd.read_json(FLIGHTS_DF_FILE_NAME).sort_index()
_pd_flights['timestamp'] = \
    pd.to_datetime(_pd_flights['timestamp'])
_pd_flights.index = _pd_flights.index.map(str)  # make index 'object' not int
_ed_flights = ed.read_es(ES_TEST_CLIENT, FLIGHTS_INDEX_NAME)

_pd_flights_small = _pd_flights.head(48)
_ed_flights_small = ed.read_es(ES_TEST_CLIENT, FLIGHTS_SMALL_INDEX_NAME)

_pd_ecommerce = pd.read_json(ECOMMERCE_DF_FILE_NAME).sort_index()
_pd_ecommerce['order_date'] = \
    pd.to_datetime(_pd_ecommerce['order_date'])
_pd_ecommerce['products.created_on'] = \
    _pd_ecommerce['products.created_on'].apply(lambda x: pd.to_datetime(x))
_pd_ecommerce.insert(2, 'customer_birth_date', None)
_pd_ecommerce.index = _pd_ecommerce.index.map(str)  # make index 'object' not int
_pd_ecommerce['customer_birth_date'].astype('datetime64')
_ed_ecommerce = ed.read_es(ES_TEST_CLIENT, ECOMMERCE_INDEX_NAME)


class TestData:

    def pd_flights(self):
        return _pd_flights

    def ed_flights(self):
        return _ed_flights

    def pd_flights_small(self):
        return _pd_flights_small

    def ed_flights_small(self):
        return _ed_flights_small

    def pd_ecommerce(self):
        return _pd_ecommerce

    def ed_ecommerce(self):
        return _ed_ecommerce


def assert_pandas_eland_frame_equal(left, right):
    if not isinstance(left, pd.DataFrame):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='pd.DataFrame', act_type=type(left)))

    if not isinstance(right, ed.DataFrame):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='ed.DataFrame', act_type=type(right)))

    # Use pandas tests to check similarity
    assert_frame_equal(left, right._to_pandas())


def assert_eland_frame_equal(left, right):
    if not isinstance(left, ed.DataFrame):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='ed.DataFrame', act_type=type(left)))

    if not isinstance(right, ed.DataFrame):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='ed.DataFrame', act_type=type(right)))

    # Use pandas tests to check similarity
    assert_frame_equal(left._to_pandas(), right._to_pandas())


def assert_pandas_eland_series_equal(left, right, check_less_precise=False):
    if not isinstance(left, pd.Series):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='pd.Series', act_type=type(left)))

    if not isinstance(right, ed.Series):
        raise AssertionError("Expected type {exp_type}, found {act_type} instead".format(
            exp_type='ed.Series', act_type=type(right)))

    # Use pandas tests to check similarity
    assert_series_equal(left, right._to_pandas(), check_less_precise=check_less_precise)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
from io import StringIO

import numpy as np

import eland as ed
from eland.tests import FLIGHTS_INDEX_NAME, ES_TEST_CLIENT
from eland.tests.common import TestData


class TestScriptedFields(TestData):

    def test_add_new_scripted_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        ed_field_mappings.add_scripted_field('scripted_field_None', None, np.dtype('int64'))

        # note 'None' is printed as 'NaN' in index, but .index shows it is 'None'
        # buf = StringIO()
        # ed_field_mappings.info_es(buf)
        # print(buf.getvalue())

        expected = self.pd_flights().columns.to_list()
        expected.append(None)

        assert expected == ed_field_mappings.display_names

    def test_add_duplicate_scripted_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        ed_field_mappings.add_scripted_field('scripted_field_Carrier', 'Carrier', np.dtype('int64'))

        # note 'None' is printed as 'NaN' in index, but .index shows it is 'None'
        buf = StringIO()
        ed_field_mappings.info_es(buf)
        print(buf.getvalue())

        expected = self.pd_flights().columns.to_list()
        expected.remove('Carrier')
        expected.append('Carrier')

        assert expected == ed_field_mappings.display_names



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import numpy as np
import pandas as pd
import pytest
from pandas.util.testing import assert_almost_equal

from eland.tests.common import TestData


class TestSeriesFrameHist(TestData):

    def test_flight_delay_min_hist(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        num_bins = 10

        # pandas data
        pd_flightdelaymin = np.histogram(pd_flights['FlightDelayMin'], num_bins)

        pd_bins = pd.DataFrame(
            {'FlightDelayMin': pd_flightdelaymin[1]})
        pd_weights = pd.DataFrame(
            {'FlightDelayMin': pd_flightdelaymin[0]})

        ed_bins, ed_weights = ed_flights['FlightDelayMin']._hist(num_bins=num_bins)

        # Numbers are slightly different
        print(pd_bins, ed_bins)
        assert_almost_equal(pd_bins, ed_bins)
        assert_almost_equal(pd_weights, ed_weights)

    def test_filtered_hist(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        num_bins = 10

        # pandas data
        pd_filteredhist = np.histogram(pd_flights[pd_flights.FlightDelay == True].FlightDelayMin, num_bins)

        pd_bins = pd.DataFrame(
            {'FlightDelayMin': pd_filteredhist[1]})
        pd_weights = pd.DataFrame(
            {'FlightDelayMin': pd_filteredhist[0]})

        d = ed_flights[ed_flights.FlightDelay == True].FlightDelayMin
        print(d.info_es())

        ed_bins, ed_weights = ed_flights[ed_flights.FlightDelay == True].FlightDelayMin._hist(num_bins=num_bins)

        # Numbers are slightly different
        assert_almost_equal(pd_bins, ed_bins)
        assert_almost_equal(pd_weights, ed_weights)

    def test_invalid_hist(self):
        with pytest.raises(ValueError):
            assert self.ed_ecommerce()['products.tax_amount'].hist()



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import numpy as np
from pandas.util.testing import assert_series_equal

from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_frame_equal


class TestDataFrameDtypes(TestData):

    def test_flights_dtypes(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        print(pd_flights.dtypes)
        print(ed_flights.dtypes)

        assert_series_equal(pd_flights.dtypes, ed_flights.dtypes)

        for i in range(0, len(pd_flights.dtypes) - 1):
            assert isinstance(pd_flights.dtypes[i], type(ed_flights.dtypes[i]))

    def test_flights_select_dtypes(self):
        pd_flights = self.pd_flights_small()
        ed_flights = self.ed_flights_small()

        assert_pandas_eland_frame_equal(
            pd_flights.select_dtypes(include=np.number),
            ed_flights.select_dtypes(include=np.number)
        )



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import warnings
from collections import OrderedDict

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import (is_float_dtype, is_bool_dtype, is_integer_dtype, is_datetime_or_timedelta_dtype,
                                       is_string_dtype)
from pandas.core.dtypes.inference import is_list_like


class FieldMappings:
    """
    General purpose to manage Elasticsearch to/from pandas mappings

    Attributes
    ----------

    _mappings_capabilities: pandas.DataFrame
        A data frame summarising the capabilities of the index mapping

        index                       - the eland display name

        es_field_name               - the Elasticsearch field name
        is_source                   - is top level field (i.e. not a multi-field sub-field)
        es_dtype                    - Elasticsearch field datatype
        es_date_format              - Elasticsearch date format (or None)
        pd_dtype                    - Pandas datatype
        is_searchable               - is the field searchable?
        is_aggregatable             - is the field aggregatable?
        is_scripted                 - is the field a scripted_field?
        aggregatable_es_field_name  - either es_field_name (if aggregatable),
                                      or es_field_name.keyword (if exists) or None
    """

    # the labels for each column (display_name is index)
    column_labels = ['es_field_name', 'is_source', 'es_dtype', 'es_date_format', 'pd_dtype', 'is_searchable',
                     'is_aggregatable', 'is_scripted', 'aggregatable_es_field_name']

    def __init__(self,
                 client=None,
                 index_pattern=None,
                 display_names=None):
        """
        Parameters
        ----------
        client: eland.Client
            Elasticsearch client

        index_pattern: str
            Elasticsearch index pattern

        display_names: list of str
            Field names to display
        """
        if (client is None) or (index_pattern is None):
            raise ValueError("Can not initialise mapping without client or index_pattern {} {}", client, index_pattern)

        get_mapping = client.get_mapping(index=index_pattern)

        # Get all fields (including all nested) and then all field_caps
        all_fields = FieldMappings._extract_fields_from_mapping(get_mapping)
        all_fields_caps = client.field_caps(index=index_pattern, fields='*')

        # Get top level (not sub-field multifield) mappings
        source_fields = FieldMappings._extract_fields_from_mapping(get_mapping, source_only=True)

        # Populate capability matrix of fields
        self._mappings_capabilities = FieldMappings._create_capability_matrix(all_fields, source_fields,
                                                                              all_fields_caps)

        if display_names is not None:
            self.display_names = display_names

    @staticmethod
    def _extract_fields_from_mapping(mappings, source_only=False, date_format=None):
        """
        Extract all field names and types from a mapping.
        ```
        {
          "my_index": {
            "mappings": {
              "properties": {
                "city": {
                  "type": "text",
                  "fields": {
                    "keyword": {
                      "type": "keyword"
                    }
                  }
                }
              }
            }
          }
        }
        ```
        if source_only == False:
            return {'city': 'text', 'city.keyword': 'keyword'}
        else:
            return {'city': 'text'}

        Note: first field name type wins. E.g.

        ```
        PUT my_index1 {"mappings":{"properties":{"city":{"type":"text"}}}}
        PUT my_index2 {"mappings":{"properties":{"city":{"type":"long"}}}}

        Returns {'city': 'text'}
        ```

        Parameters
        ----------
        mappings: dict
            Return from get_mapping

        Returns
        -------
        fields, dates_format: tuple(OrderedDict, dict)
            where:
                fields: OrderedDict of field names and types
                dates_format: Dict of date field names and format

        """
        fields = OrderedDict()

        # Recurse until we get a 'type: xxx'
        def flatten(x, name=''):
            if type(x) is dict:
                for a in x:
                    if a == 'type' and type(x[a]) is str:  # 'type' can be a name of a field
                        field_name = name[:-1]
                        field_type = x[a]
                        # if field_type is 'date' keep track of the format info when available
                        date_format = None
                        if field_type == "date" and "format" in x:
                            date_format = x["format"]
                        # If there is a conflicting type, warn - first values added wins
                        if field_name in fields and fields[field_name] != field_type:
                            warnings.warn("Field {} has conflicting types {} != {}".
                                          format(field_name, fields[field_name], field_type),
                                          UserWarning)
                        else:
                            fields[field_name] = (field_type, date_format)
                    elif a == 'properties' or (not source_only and a == 'fields'):
                        flatten(x[a], name)
                    elif not (source_only and a == 'fields'):  # ignore multi-field fields for source_only
                        flatten(x[a], name + a + '.')

        for index in mappings:
            if 'properties' in mappings[index]['mappings']:
                properties = mappings[index]['mappings']['properties']

                flatten(properties)

        return fields

    @staticmethod
    def _create_capability_matrix(all_fields, source_fields, all_fields_caps):
        """
        {
          "fields": {
            "rating": {
              "long": {
                "searchable": true,
                "aggregatable": false,
                "indices": ["index1", "index2"],
                "non_aggregatable_indices": ["index1"]
              },
              "keyword": {
                "searchable": false,
                "aggregatable": true,
                "indices": ["index3", "index4"],
                "non_searchable_indices": ["index4"]
              }
            },
            "title": {
              "text": {
                "searchable": true,
                "aggregatable": false

              }
            }
          }
        }
        """
        all_fields_caps_fields = all_fields_caps['fields']

        capability_matrix = OrderedDict()

        for field, field_caps in all_fields_caps_fields.items():
            if field in all_fields:
                # v = {'long': {'type': 'long', 'searchable': True, 'aggregatable': True}}
                for kk, vv in field_caps.items():
                    _source = (field in source_fields)
                    es_field_name = field
                    es_dtype = vv['type']
                    es_date_format = all_fields[field][1]
                    pd_dtype = FieldMappings._es_dtype_to_pd_dtype(vv['type'])
                    is_searchable = vv['searchable']
                    is_aggregatable = vv['aggregatable']
                    scripted = False
                    aggregatable_es_field_name = None  # this is populated later

                    caps = [es_field_name, _source, es_dtype, es_date_format, pd_dtype, is_searchable, is_aggregatable,
                            scripted, aggregatable_es_field_name]

                    capability_matrix[field] = caps

                    if 'non_aggregatable_indices' in vv:
                        warnings.warn("Field {} has conflicting aggregatable fields across indexes {}",
                                      format(field, vv['non_aggregatable_indices']),
                                      UserWarning)
                    if 'non_searchable_indices' in vv:
                        warnings.warn("Field {} has conflicting searchable fields across indexes {}",
                                      format(field, vv['non_searchable_indices']),
                                      UserWarning)

        capability_matrix_df = pd.DataFrame.from_dict(capability_matrix, orient='index',
                                                      columns=FieldMappings.column_labels)

        def find_aggregatable(row, df):
            # convert series to dict so we can add 'aggregatable_es_field_name'
            row_as_dict = row.to_dict()
            if row_as_dict['is_aggregatable'] == False:
                # if not aggregatable, then try field.keyword
                es_field_name_keyword = row.es_field_name + '.keyword'
                try:
                    series = df.loc[df.es_field_name == es_field_name_keyword]
                    if not series.empty and series.is_aggregatable.squeeze():
                        row_as_dict['aggregatable_es_field_name'] = es_field_name_keyword
                    else:
                        row_as_dict['aggregatable_es_field_name'] = None
                except KeyError:
                    row_as_dict['aggregatable_es_field_name'] = None
            else:
                row_as_dict['aggregatable_es_field_name'] = row_as_dict['es_field_name']

            return pd.Series(data=row_as_dict)

        # add aggregatable_es_field_name column by applying action to each row
        capability_matrix_df = capability_matrix_df.apply(find_aggregatable, args=(capability_matrix_df,),
                                                          axis='columns')

        # return just source fields (as these are the only ones we display)
        return capability_matrix_df[capability_matrix_df.is_source].sort_index()

    @staticmethod
    def _es_dtype_to_pd_dtype(es_dtype):
        """
        Mapping Elasticsearch types to pandas dtypes
        --------------------------------------------

        Elasticsearch field datatype              | Pandas dtype
        --
        text                                      | object
        keyword                                   | object
        long, integer, short, byte, binary        | int64
        double, float, half_float, scaled_float   | float64
        date, date_nanos                          | datetime64
        boolean                                   | bool
        TODO - add additional mapping types
        """
        es_dtype_to_pd_dtype = {
            'text': 'object',
            'keyword': 'object',

            'long': 'int64',
            'integer': 'int64',
            'short': 'int64',
            'byte': 'int64',
            'binary': 'int64',

            'double': 'float64',
            'float': 'float64',
            'half_float': 'float64',
            'scaled_float': 'float64',

            'date': 'datetime64[ns]',
            'date_nanos': 'datetime64[ns]',

            'boolean': 'bool'
        }

        if es_dtype in es_dtype_to_pd_dtype:
            return es_dtype_to_pd_dtype[es_dtype]

        # Return 'object' for all unsupported TODO - investigate how different types could be supported
        return 'object'

    @staticmethod
    def _pd_dtype_to_es_dtype(pd_dtype):
        """
        Mapping pandas dtypes to Elasticsearch dtype
        --------------------------------------------

        ```
        Pandas dtype	Python type	NumPy type	Usage
        object	str	string_, unicode_	Text
        int64	int	int_, int8, int16, int32, int64, uint8, uint16, uint32, uint64	Integer numbers
        float64	float	float_, float16, float32, float64	Floating point numbers
        bool	bool	bool_	True/False values
        datetime64	NA	datetime64[ns]	Date and time values
        timedelta[ns]	NA	NA	Differences between two datetimes
        category	NA	NA	Finite list of text values
        ```
        """
        es_dtype = None

        # Map all to 64-bit - TODO map to specifics: int32 -> int etc.
        if is_float_dtype(pd_dtype):
            es_dtype = 'double'
        elif is_integer_dtype(pd_dtype):
            es_dtype = 'long'
        elif is_bool_dtype(pd_dtype):
            es_dtype = 'boolean'
        elif is_string_dtype(pd_dtype):
            es_dtype = 'keyword'
        elif is_datetime_or_timedelta_dtype(pd_dtype):
            es_dtype = 'date'
        else:
            warnings.warn('No mapping for pd_dtype: [{0}], using default mapping'.format(pd_dtype))

        return es_dtype

    @staticmethod
    def _generate_es_mappings(dataframe, geo_points=None):
        """Given a pandas dataframe, generate the associated Elasticsearch mapping

        Parameters
        ----------
            dataframe : pandas.DataFrame
                pandas.DataFrame to create schema from

        Returns
        -------
            mapping : str
        """

        """
        "mappings" : {
          "properties" : {
            "AvgTicketPrice" : {
              "type" : "float"
            },
            "Cancelled" : {
              "type" : "boolean"
            },
            "Carrier" : {
              "type" : "keyword"
            },
            "Dest" : {
              "type" : "keyword"
            }
          }
        }
        """

        mappings = {'properties': {}}
        for field_name_name, dtype in dataframe.dtypes.iteritems():
            if geo_points is not None and field_name_name in geo_points:
                es_dtype = 'geo_point'
            else:
                es_dtype = FieldMappings._pd_dtype_to_es_dtype(dtype)

            mappings['properties'][field_name_name] = OrderedDict()
            mappings['properties'][field_name_name]['type'] = es_dtype

        return {"mappings": mappings}

    def aggregatable_field_name(self, display_name):
        """
        Return a single aggregatable field_name from display_name

        Logic here is that field_name names are '_source' fields and keyword fields
        may be nested beneath the field. E.g.
        customer_full_name: text
        customer_full_name.keyword: keyword

        customer_full_name.keyword is the aggregatable field for customer_full_name

        Parameters
        ----------
        display_name: str

        Returns
        -------
        aggregatable_es_field_name: str or None
            The aggregatable field name associated with display_name. This could be the field_name, or the
            field_name.keyword.

        raise KeyError if the field_name doesn't exist in the mapping, or isn't aggregatable
        """
        if display_name not in self._mappings_capabilities.index:
            raise KeyError("Can not get aggregatable field name for invalid display name {}".format(display_name))

        if self._mappings_capabilities.loc[display_name].aggregatable_es_field_name is None:
            warnings.warn("Aggregations not supported for '{}' '{}'".format(display_name,
                                                                            self._mappings_capabilities.loc[
                                                                                display_name].es_field_name))

        return self._mappings_capabilities.loc[display_name].aggregatable_es_field_name

    def aggregatable_field_names(self):
        """
        Return a list of aggregatable Elasticsearch field_names for all display names.
        If field is not aggregatable_field_names, return nothing.

        Logic here is that field_name names are '_source' fields and keyword fields
        may be nested beneath the field. E.g.
        customer_full_name: text
        customer_full_name.keyword: keyword

        customer_full_name.keyword is the aggregatable field for customer_full_name

        Returns
        -------
        Dict of aggregatable_field_names
            key = aggregatable_field_name, value = field_name
            e.g. {'customer_full_name.keyword': 'customer_full_name', ...}
        """
        non_aggregatables = self._mappings_capabilities[self._mappings_capabilities.aggregatable_es_field_name.isna()]
        if not non_aggregatables.empty:
            warnings.warn("Aggregations not supported for '{}'".format(non_aggregatables))

        aggregatables = self._mappings_capabilities[self._mappings_capabilities.aggregatable_es_field_name.notna()]

        # extract relevant fields and convert to dict
        # <class 'dict'>: {'category.keyword': 'category', 'currency': 'currency', ...
        return OrderedDict(
            aggregatables[['aggregatable_es_field_name', 'es_field_name']].to_dict(orient='split')['data'])

    def date_field_format(self, es_field_name):
        """
        Parameters
        ----------
        es_field_name: str


        Returns
        -------
        str
            A string (for date fields) containing the date format for the field
        """
        return self._mappings_capabilities.loc[
            self._mappings_capabilities.es_field_name == es_field_name].es_date_format.squeeze()

    def field_name_pd_dtype(self, es_field_name):
        """
        Parameters
        ----------
        es_field_name: str

        Returns
        -------
        pd_dtype: str
            The pandas data type we map to

        Raises
        ------
        KeyError
            If es_field_name does not exist in mapping
        """
        if es_field_name not in self._mappings_capabilities.es_field_name:
            raise KeyError("es_field_name {} does not exist".format(es_field_name))

        pd_dtype = self._mappings_capabilities.loc[
            self._mappings_capabilities.es_field_name == es_field_name
            ].pd_dtype.squeeze()
        return pd_dtype

    def add_scripted_field(self, scripted_field_name, display_name, pd_dtype):
        # if this display name is used somewhere else, drop it
        if display_name in self._mappings_capabilities.index:
            self._mappings_capabilities = self._mappings_capabilities.drop(index=[display_name])

        # ['es_field_name', 'is_source', 'es_dtype', 'es_date_format', 'pd_dtype', 'is_searchable',
        # 'is_aggregatable', 'is_scripted', 'aggregatable_es_field_name']

        capabilities = {display_name: [scripted_field_name,
                                       False,
                                       self._pd_dtype_to_es_dtype(pd_dtype),
                                       None,
                                       pd_dtype,
                                       True,
                                       True,
                                       True,
                                       scripted_field_name]}

        capability_matrix_row = pd.DataFrame.from_dict(capabilities, orient='index',
                                                       columns=FieldMappings.column_labels)

        self._mappings_capabilities = self._mappings_capabilities.append(capability_matrix_row)

    def numeric_source_fields(self):
        pd_dtypes, es_field_names, es_date_formats = self.metric_source_fields()
        return es_field_names

    def metric_source_fields(self, include_bool=False, include_timestamp=False):
        """
        Returns
        -------
        pd_dtypes: list of np.dtype
            List of pd_dtypes for es_field_names
        es_field_names: list of str
            List of source fields where pd_dtype == (int64 or float64 or bool or timestamp)
        es_date_formats: list of str (can be None)
            List of es date formats for es_field

        TODO - not very efficient, but unless called per row, this should be ok
        """
        pd_dtypes = []
        es_field_names = []
        es_date_formats = []
        for index, row in self._mappings_capabilities.iterrows():
            pd_dtype = row['pd_dtype']
            es_field_name = row['es_field_name']
            es_date_format = row['es_date_format']

            if is_integer_dtype(pd_dtype) or is_float_dtype(pd_dtype):
                pd_dtypes.append(np.dtype(pd_dtype))
                es_field_names.append(es_field_name)
                es_date_formats.append(es_date_format)
            elif include_bool and is_bool_dtype(pd_dtype):
                pd_dtypes.append(np.dtype(pd_dtype))
                es_field_names.append(es_field_name)
                es_date_formats.append(es_date_format)
            elif include_timestamp and is_datetime_or_timedelta_dtype(pd_dtype):
                pd_dtypes.append(np.dtype(pd_dtype))
                es_field_names.append(es_field_name)
                es_date_formats.append(es_date_format)

        # return in display_name order
        return pd_dtypes, es_field_names, es_date_formats

    def get_field_names(self, include_scripted_fields=True):
        if include_scripted_fields:
            return self._mappings_capabilities.es_field_name.to_list()

        return self._mappings_capabilities[
            self._mappings_capabilities.is_scripted == False
            ].es_field_name.to_list()

    def _get_display_names(self):
        return self._mappings_capabilities.index.to_list()

    def _set_display_names(self, display_names):
        if not is_list_like(display_names):
            raise ValueError("'{}' is not list like".format(display_names))

        if list(set(display_names) - set(self.display_names)):
            raise KeyError("{} not in display names {}".format(display_names, self.display_names))

        self._mappings_capabilities = self._mappings_capabilities.reindex(display_names)

    display_names = property(_get_display_names, _set_display_names)

    def dtypes(self):
        """
        Returns
        -------
        dtypes: pd.Series
            Index: Display name
            Values: pd_dtype as np.dtype
        """
        pd_dtypes = self._mappings_capabilities['pd_dtype']

        # Set name of the returned series as None
        pd_dtypes.name = None

        # Convert return from 'str' to 'np.dtype'
        return pd_dtypes.apply(lambda x: np.dtype(x))

    def info_es(self, buf):
        buf.write("Mappings:\n")
        buf.write(" capabilities:\n{0}\n".format(self._mappings_capabilities.to_string()))

    def rename(self, old_name_new_name_dict):
        """
        Renames display names in-place

        Parameters
        ----------
        old_name_new_name_dict

        Returns
        -------
        Nothing

        Notes
        -----
        For the names that do not exist this is a no op
        """
        self._mappings_capabilities = self._mappings_capabilities.rename(index=old_name_new_name_dict)

    def get_renames(self):
        # return dict of renames { old_name: new_name, ... } (inefficient)
        renames = {}

        for display_name in self.display_names:
            field_name = self._mappings_capabilities.loc[display_name].es_field_name
            if field_name != display_name:
                renames[field_name] = display_name

        return renames



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from eland.tests.common import TestData
from eland.tests.common import (
    assert_pandas_eland_frame_equal
)


class TestDataFrameDrop(TestData):

    def test_flights_small_drop(self):
        ed_flights_small = self.ed_flights_small()
        pd_flights_small = self.pd_flights_small()

        # ['AvgTicketPrice', 'Cancelled', 'Carrier', 'Dest', 'DestAirportID',
        #        'DestCityName', 'DestCountry', 'DestLocation', 'DestRegion',
        #        'DestWeather', 'DistanceKilometers', 'DistanceMiles', 'FlightDelay',
        #        'FlightDelayMin', 'FlightDelayType', 'FlightNum', 'FlightTimeHour',
        #        'FlightTimeMin', 'Origin', 'OriginAirportID', 'OriginCityName',
        #        'OriginCountry', 'OriginLocation', 'OriginRegion', 'OriginWeather',
        #        'dayOfWeek', 'timestamp']
        pd_col0 = pd_flights_small.drop(['Carrier', 'DestCityName'], axis=1)
        pd_col1 = pd_flights_small.drop(columns=['Carrier', 'DestCityName'])

        ed_col0 = ed_flights_small.drop(['Carrier', 'DestCityName'], axis=1)
        ed_col1 = ed_flights_small.drop(columns=['Carrier', 'DestCityName'])

        assert_pandas_eland_frame_equal(pd_col0, ed_col0)
        assert_pandas_eland_frame_equal(pd_col1, ed_col1)

        # Drop rows by index
        pd_idx0 = pd_flights_small.drop(['1', '2'])
        ed_idx0 = ed_flights_small.drop(['1', '2'])

        assert_pandas_eland_frame_equal(pd_idx0, ed_idx0)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import eland as ed
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_INDEX_NAME
from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_series_equal


class TestSeriesRename(TestData):

    def test_rename(self):
        pd_carrier = self.pd_flights()['Carrier']
        ed_carrier = ed.Series(ES_TEST_CLIENT, FLIGHTS_INDEX_NAME, 'Carrier')

        assert_pandas_eland_series_equal(pd_carrier, ed_carrier)

        pd_renamed = pd_carrier.rename("renamed")
        ed_renamed = ed_carrier.rename("renamed")

        print(pd_renamed)
        print(ed_renamed)

        print(ed_renamed.info_es())

        assert_pandas_eland_series_equal(pd_renamed, ed_renamed)

        pd_renamed2 = pd_renamed.rename("renamed2")
        ed_renamed2 = ed_renamed.rename("renamed2")

        print(ed_renamed2.info_es())

        assert "renamed2" == ed_renamed2.name

        assert_pandas_eland_series_equal(pd_renamed2, ed_renamed2)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import numpy as np
import pandas as pd

import eland as ed
from eland.tests.common import ES_TEST_CLIENT, assert_pandas_eland_frame_equal
from eland.tests.common import TestData


class TestDataFrameUtils(TestData):

    def test_generate_es_mappings(self):
        df = pd.DataFrame(data={'A': np.random.rand(3),
                                'B': 1,
                                'C': 'foo',
                                'D': pd.Timestamp('20190102'),
                                'E': [1.0, 2.0, 3.0],
                                'F': False,
                                'G': [1, 2, 3]},
                          index=['0', '1', '2'])

        expected_mappings = {'mappings': {
            'properties': {'A': {'type': 'double'},
                           'B': {'type': 'long'},
                           'C': {'type': 'keyword'},
                           'D': {'type': 'date'},
                           'E': {'type': 'double'},
                           'F': {'type': 'boolean'},
                           'G': {'type': 'long'}}}}

        mappings = ed.FieldMappings._generate_es_mappings(df)

        assert expected_mappings == mappings

        # Now create index
        index_name = 'eland_test_generate_es_mappings'

        ed_df = ed.pandas_to_eland(df, ES_TEST_CLIENT, index_name, es_if_exists="replace", es_refresh=True)
        ed_df_head = ed_df.head()

        assert_pandas_eland_frame_equal(df, ed_df_head)

    def test_eland_to_pandas_performance(self):
        # TODO quantify this
        pd_df = ed.eland_to_pandas(self.ed_flights(), show_progress=True)

        # This test calls the same method so is redundant
        #assert_pandas_eland_frame_equal(pd_df, self.ed_flights())



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import warnings
from copy import deepcopy

from eland.filter import BooleanFilter, NotNull, IsNull, IsIn


class Query:
    """
    Simple class to manage building Elasticsearch queries.
    """

    def __init__(self, query=None):
        if query is None:
            self._query = BooleanFilter()
            self._aggs = {}
        else:
            # Deep copy the incoming query so we can change it
            self._query = deepcopy(query._query)
            self._aggs = deepcopy(query._aggs)

    def exists(self, field, must=True):
        """
        Add exists query
        https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-exists-query.html
        """
        if must:
            if self._query.empty():
                self._query = NotNull(field)
            else:
                self._query = self._query & NotNull(field)
        else:
            if self._query.empty():
                self._query = IsNull(field)
            else:
                self._query = self._query & IsNull(field)

    def ids(self, items, must=True):
        """
        Add ids query
        https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-ids-query.html
        """
        if must:
            if self._query.empty():
                self._query = IsIn('ids', items)
            else:
                self._query = self._query & IsIn('ids', items)
        else:
            if self._query.empty():
                self._query = ~(IsIn('ids', items))
            else:
                self._query = self._query & ~(IsIn('ids', items))

    def terms(self, field, items, must=True):
        """
        Add ids query
        https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-terms-query.html
        """
        if must:
            if self._query.empty():
                self._query = IsIn(field, items)
            else:
                self._query = self._query & IsIn(field, items)
        else:
            if self._query.empty():
                self._query = ~(IsIn(field, items))
            else:
                self._query = self._query & ~(IsIn(field, items))

    def terms_aggs(self, name, func, field, es_size):
        """
        Add terms agg e.g

        "aggs": {
            "name": {
                "terms": {
                    "field": "Airline",
                    "size": 10
                }
            }
        }
        """
        agg = {
            func: {
                "field": field,
                "size": es_size
            }
        }
        self._aggs[name] = agg

    def metric_aggs(self, name, func, field):
        """
        Add metric agg e.g

        "aggs": {
            "name": {
                "max": {
                    "field": "AvgTicketPrice"
                }
            }
        }
        """
        agg = {
            func: {
                "field": field
            }
        }
        self._aggs[name] = agg

    def hist_aggs(self, name, field, min_aggs, max_aggs, num_bins):
        """
        Add histogram agg e.g.
        "aggs": {
            "name": {
                "histogram": {
                    "field": "AvgTicketPrice"
                    "interval": (max_aggs[field] - min_aggs[field])/bins
                }
            }
        }
        """
        min = min_aggs[field]
        max = max_aggs[field]

        interval = (max - min) / num_bins
        offset = min

        agg = {
            "histogram": {
                "field": field,
                "interval": interval,
                "offset": offset
            }
        }

        if interval != 0:
            self._aggs[name] = agg

    def to_search_body(self):
        if self._query.empty():
            if self._aggs:
                body = {"aggs": self._aggs}
            else:
                body = {}
        else:
            if self._aggs:
                body = {"query": self._query.build(), "aggs": self._aggs}
            else:
                body = {"query": self._query.build()}
        return body

    def to_count_body(self):
        if len(self._aggs) > 0:
            warnings.warn('Requesting count for agg query {}', self)
        if self._query.empty():
            body = None
        else:
            body = {"query": self._query.build()}

        return body

    def update_boolean_filter(self, boolean_filter):
        if self._query.empty():
            self._query = boolean_filter
        else:
            self._query = self._query & boolean_filter

    def __repr__(self):
        return repr(self.to_search_body())



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import numpy as np
from pandas.util.testing import assert_almost_equal

from eland.tests.common import TestData


class TestDataFrameAggs(TestData):

    def test_basic_aggs(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        pd_sum_min = pd_flights.select_dtypes(include=[np.number]).agg(['sum', 'min'])
        ed_sum_min = ed_flights.select_dtypes(include=[np.number]).agg(['sum', 'min'])

        # Eland returns all float values for all metric aggs, pandas can return int
        # TODO - investigate this more
        pd_sum_min = pd_sum_min.astype('float64')
        assert_almost_equal(pd_sum_min, ed_sum_min)

        pd_sum_min_std = pd_flights.select_dtypes(include=[np.number]).agg(['sum', 'min', 'std'])
        ed_sum_min_std = ed_flights.select_dtypes(include=[np.number]).agg(['sum', 'min', 'std'])

        print(pd_sum_min_std.dtypes)
        print(ed_sum_min_std.dtypes)

        assert_almost_equal(pd_sum_min_std, ed_sum_min_std, check_less_precise=True)

    def test_terms_aggs(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        pd_sum_min = pd_flights.select_dtypes(include=[np.number]).agg(['sum', 'min'])
        ed_sum_min = ed_flights.select_dtypes(include=[np.number]).agg(['sum', 'min'])

        # Eland returns all float values for all metric aggs, pandas can return int
        # TODO - investigate this more
        pd_sum_min = pd_sum_min.astype('float64')
        assert_almost_equal(pd_sum_min, ed_sum_min)

        pd_sum_min_std = pd_flights.select_dtypes(include=[np.number]).agg(['sum', 'min', 'std'])
        ed_sum_min_std = ed_flights.select_dtypes(include=[np.number]).agg(['sum', 'min', 'std'])

        print(pd_sum_min_std.dtypes)
        print(ed_sum_min_std.dtypes)

        assert_almost_equal(pd_sum_min_std, ed_sum_min_std, check_less_precise=True)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from elasticsearch import Elasticsearch
from elasticsearch import helpers


class Client:
    """
    eland client - implemented as facade to control access to Elasticsearch methods
    """

    def __init__(self, es=None):
        if isinstance(es, Elasticsearch):
            self._es = es
        elif isinstance(es, Client):
            self._es = es._es
        else:
            self._es = Elasticsearch(es)

    def index_create(self, **kwargs):
        return self._es.indices.create(**kwargs)

    def index_delete(self, **kwargs):
        return self._es.indices.delete(**kwargs)

    def index_exists(self, **kwargs):
        return self._es.indices.exists(**kwargs)

    def get_mapping(self, **kwargs):
        return self._es.indices.get_mapping(**kwargs)

    def bulk(self, actions, refresh=False):
        return helpers.bulk(self._es, actions, refresh=refresh)

    def scan(self, **kwargs):
        return helpers.scan(self._es, **kwargs)

    def search(self, **kwargs):
        return self._es.search(**kwargs)

    def field_caps(self, **kwargs):
        return self._es.field_caps(**kwargs)

    def count(self, **kwargs):
        count_json = self._es.count(**kwargs)
        return count_json['count']



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import pytest
from pandas.util.testing import assert_series_equal

from eland.tests.common import TestData


class TestSeriesValueCounts(TestData):

    def test_value_counts(self):
        pd_s = self.pd_flights()['Carrier']
        ed_s = self.ed_flights()['Carrier']

        pd_vc = pd_s.value_counts()
        ed_vc = ed_s.value_counts()

        assert_series_equal(pd_vc, ed_vc)

    def test_value_counts_size(self):
        pd_s = self.pd_flights()['Carrier']
        ed_s = self.ed_flights()['Carrier']

        pd_vc = pd_s.value_counts()[:1]
        ed_vc = ed_s.value_counts(es_size=1)

        assert_series_equal(pd_vc, ed_vc)

    def test_value_counts_keyerror(self):
        ed_f = self.ed_flights()
        with pytest.raises(KeyError):
            assert ed_f['not_a_column'].value_counts()

    def test_value_counts_dataframe(self):
        # value_counts() is a series method, should raise AttributeError if called on a DataFrame
        ed_f = self.ed_flights()
        with pytest.raises(AttributeError):
            assert ed_f.value_counts()

    def test_value_counts_non_int(self):
        ed_s = self.ed_flights()['Carrier']
        with pytest.raises(TypeError):
            assert ed_s.value_counts(es_size='foo')

    def test_value_counts_non_positive_int(self):
        ed_s = self.ed_flights()['Carrier']
        with pytest.raises(ValueError):
            assert ed_s.value_counts(es_size=-9)

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_value_counts_non_aggregatable(self):
        ed_s = self.ed_ecommerce()['customer_first_name']
        pd_s = self.pd_ecommerce()['customer_first_name']

        pd_vc = pd_s.value_counts().head(20).sort_index()
        ed_vc = ed_s.value_counts(es_size=20).sort_index()

        assert_series_equal(pd_vc, ed_vc)

        ed_s = self.ed_ecommerce()['customer_gender']
        with pytest.raises(ValueError):
            assert ed_s.value_counts()



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
from io import StringIO

from eland.tests.common import TestData


class TestDataFrameInfo(TestData):

    def test_flights_info(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_buf = StringIO()
        pd_buf = StringIO()

        # Ignore memory_usage and first line (class name)
        ed_flights.info(buf=ed_buf, memory_usage=False)
        pd_flights.info(buf=pd_buf, memory_usage=False)

        ed_buf_lines = ed_buf.getvalue().split('\n')
        pd_buf_lines = pd_buf.getvalue().split('\n')

        assert pd_buf_lines[1:] == ed_buf_lines[1:]

        # NOTE: info does not work on truncated data frames (e.g. head/tail) TODO

        print(self.ed_ecommerce().info())



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import numpy as np
import pandas as pd
import pytest

import eland as ed

# Fix console size for consistent test results
pd.set_option('display.max_rows', 10)
pd.set_option('display.max_columns', 5)
pd.set_option('display.width', 100)


@pytest.fixture(autouse=True)
def add_imports(doctest_namespace):
    doctest_namespace["np"] = np
    doctest_namespace["pd"] = pd
    doctest_namespace["ed"] = ed



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.util.testing import assert_almost_equal

from eland.tests.common import TestData


class TestDataFrameDescribe(TestData):

    def test_flights_describe(self):
        pd_flights = self.pd_flights()
        ed_flights = self.ed_flights()

        pd_describe = pd_flights.describe()
        ed_describe = ed_flights.describe()

        assert_almost_equal(pd_describe.drop(['25%', '50%', '75%'], axis='index'),
                            ed_describe.drop(['25%', '50%', '75%'], axis='index'),
                            check_less_precise=True)

        # TODO - this fails for percentile fields as ES aggregations are approximate
        #        if ES percentile agg uses
        #        "hdr": {
        #           "number_of_significant_value_digits": 3
        #         }
        #        this works

        # pd_ecommerce_describe = self.pd_ecommerce().describe()
        # ed_ecommerce_describe = self.ed_ecommerce().describe()
        # We don't compare ecommerce here as the default dtypes in pandas from read_json
        # don't match the mapping types. This is mainly because the products field is
        # nested and so can be treated as a multi-field in ES, but not in pandas

        # We can not also run 'describe' on a truncate ed dataframe



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_frame_equal


class TestDataFrameHeadTail(TestData):

    def test_head(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head_10 = ed_flights.head(10)
        pd_head_10 = pd_flights.head(10)
        assert_pandas_eland_frame_equal(pd_head_10, ed_head_10)

        ed_head_8 = ed_head_10.head(8)
        pd_head_8 = pd_head_10.head(8)
        assert_pandas_eland_frame_equal(pd_head_8, ed_head_8)

        ed_head_20 = ed_head_10.head(20)
        pd_head_20 = pd_head_10.head(20)
        assert_pandas_eland_frame_equal(pd_head_20, ed_head_20)

    def test_tail(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_tail_10 = ed_flights.tail(10)
        pd_tail_10 = pd_flights.tail(10)
        assert_pandas_eland_frame_equal(pd_tail_10, ed_tail_10)

        ed_tail_8 = ed_tail_10.tail(8)
        pd_tail_8 = pd_tail_10.tail(8)
        assert_pandas_eland_frame_equal(pd_tail_8, ed_tail_8)

        ed_tail_20 = ed_tail_10.tail(20)
        pd_tail_20 = pd_tail_10.tail(20)
        assert_pandas_eland_frame_equal(pd_tail_20, ed_tail_20)

    def test_head_tail(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head_10 = ed_flights.head(10)
        pd_head_10 = pd_flights.head(10)
        assert_pandas_eland_frame_equal(pd_head_10, ed_head_10)

        ed_tail_8 = ed_head_10.tail(8)
        pd_tail_8 = pd_head_10.tail(8)
        assert_pandas_eland_frame_equal(pd_tail_8, ed_tail_8)

        ed_tail_5 = ed_tail_8.tail(5)
        pd_tail_5 = pd_tail_8.tail(5)
        assert_pandas_eland_frame_equal(pd_tail_5, ed_tail_5)

        ed_tail_4 = ed_tail_5.tail(4)
        pd_tail_4 = pd_tail_5.tail(4)
        assert_pandas_eland_frame_equal(pd_tail_4, ed_tail_4)

    def test_tail_head(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_tail_10 = ed_flights.tail(10)
        pd_tail_10 = pd_flights.tail(10)
        assert_pandas_eland_frame_equal(pd_tail_10, ed_tail_10)

        ed_head_8 = ed_tail_10.head(8)
        pd_head_8 = pd_tail_10.head(8)
        assert_pandas_eland_frame_equal(pd_head_8, ed_head_8)

        ed_tail_5 = ed_head_8.tail(5)
        pd_tail_5 = pd_head_8.tail(5)
        assert_pandas_eland_frame_equal(pd_tail_5, ed_tail_5)

        ed_head_4 = ed_tail_5.head(4)
        pd_head_4 = pd_tail_5.head(4)
        assert_pandas_eland_frame_equal(pd_head_4, ed_head_4)

    def test_head_0(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head_0 = ed_flights.head(0)
        pd_head_0 = pd_flights.head(0)
        assert_pandas_eland_frame_equal(pd_head_0, ed_head_0)

    def test_doc_test_tail(self):
        df = self.ed_flights()
        df = df[(df.OriginAirportID == 'AMS') & (df.FlightDelayMin > 60)]
        df = df[['timestamp', 'OriginAirportID', 'DestAirportID', 'FlightDelayMin']]
        df = df.tail()
        print(df)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
about = {}
with open(path.join(here, 'eland', '_version.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: Apache Software License",
    "Environment :: Console",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
]

LONG_DESCRIPTION="""
# What is it?

eland is a Elasticsearch client Python package to analyse, explore and manipulate data that resides in Elasticsearch. 
Where possible the package uses existing Python APIs and data structures to make it easy to switch between numpy, 
pandas, scikit-learn to their Elasticsearch powered equivalents. In general, the data resides in Elasticsearch and 
not in memory, which allows eland to access large datasets stored in Elasticsearch.

For example, to explore data in a large Elasticsearch index, simply create an eland DataFrame from an Elasticsearch 
index pattern, and explore using an API that mirrors a subset of the pandas.DataFrame API: 

```
>>> import eland as ed

>>> # Connect to 'flights' index via localhost Elasticsearch node
>>> df = ed.DataFrame('localhost:9200', 'flights') 

>>> df.head()
   AvgTicketPrice  Cancelled           Carrier  ...        OriginWeather dayOfWeek           timestamp
0      841.265642      False   Kibana Airlines  ...                Sunny         0 2018-01-01 00:00:00
1      882.982662      False  Logstash Airways  ...                Clear         0 2018-01-01 18:27:00
2      190.636904      False  Logstash Airways  ...                 Rain         0 2018-01-01 17:11:14
3      181.694216       True   Kibana Airlines  ...  Thunder & Lightning         0 2018-01-01 10:33:28
4      730.041778      False   Kibana Airlines  ...        Damaging Wind         0 2018-01-01 05:13:00

[5 rows x 27 columns]

>>> df.describe()
       AvgTicketPrice  DistanceKilometers  DistanceMiles  FlightDelayMin  FlightTimeHour  FlightTimeMin     dayOfWeek
count    13059.000000        13059.000000   13059.000000    13059.000000    13059.000000   13059.000000  13059.000000
mean       628.253689         7092.142457    4406.853010       47.335171        8.518797     511.127842      2.835975
std        266.386661         4578.263193    2844.800855       96.743006        5.579019     334.741135      1.939365
min        100.020531            0.000000       0.000000        0.000000        0.000000       0.000000      0.000000
25%        410.008918         2470.545974    1535.126118        0.000000        4.194976     251.738513      1.000000
50%        640.362667         7612.072403    4729.922470        0.000000        8.385816     503.148975      3.000000
75%        842.254990         9735.082407    6049.459005       15.000000       12.009396     720.534532      4.141221
max       1199.729004        19881.482422   12353.780273      360.000000       31.715034    1902.901978      6.000000

>>> df[['Carrier', 'AvgTicketPrice', 'Cancelled']]
                Carrier  AvgTicketPrice  Cancelled
0       Kibana Airlines      841.265642      False
1      Logstash Airways      882.982662      False
2      Logstash Airways      190.636904      False
3       Kibana Airlines      181.694216       True
4       Kibana Airlines      730.041778      False
...                 ...             ...        ...
13054  Logstash Airways     1080.446279      False
13055  Logstash Airways      646.612941      False
13056  Logstash Airways      997.751876      False
13057          JetBeats     1102.814465      False
13058          JetBeats      858.144337      False

[13059 rows x 3 columns]

>>> df[(df.Carrier=="Kibana Airlines") & (df.AvgTicketPrice > 900.0) & (df.Cancelled == True)].head()
     AvgTicketPrice  Cancelled          Carrier  ...        OriginWeather dayOfWeek           timestamp
8        960.869736       True  Kibana Airlines  ...            Heavy Fog         0 2018-01-01 12:09:35
26       975.812632       True  Kibana Airlines  ...                 Rain         0 2018-01-01 15:38:32
311      946.358410       True  Kibana Airlines  ...            Heavy Fog         0 2018-01-01 11:51:12
651      975.383864       True  Kibana Airlines  ...                 Rain         2 2018-01-03 21:13:17
950      907.836523       True  Kibana Airlines  ...  Thunder & Lightning         2 2018-01-03 05:14:51

[5 rows x 27 columns]

>>> df[['DistanceKilometers', 'AvgTicketPrice']].aggregate(['sum', 'min', 'std'])
     DistanceKilometers  AvgTicketPrice
sum        9.261629e+07    8.204365e+06
min        0.000000e+00    1.000205e+02
std        4.578263e+03    2.663867e+02

>>> df[['Carrier', 'Origin', 'Dest']].nunique()
Carrier      4
Origin     156
Dest       156
dtype: int64

>>> s = df.AvgTicketPrice * 2 + df.DistanceKilometers - df.FlightDelayMin
>>> s
0        18174.857422
1        10589.365723
2          381.273804
3          739.126221
4        14818.327637
             ...     
13054    10219.474121
13055     8381.823975
13056    12661.157104
13057    20819.488281
13058    18315.431274
Length: 13059, dtype: float64

>>> print(s.info_es())
index_pattern: flights
Index:
 index_field: _id
 is_source_field: False
Mappings:
 capabilities:
         es_field_name  is_source es_dtype es_date_format pd_dtype  is_searchable  is_aggregatable  is_scripted aggregatable_es_field_name
NaN  script_field_None      False   double           None  float64           True             True         True          script_field_None
Operations:
 tasks: []
 size: None
 sort_params: None
 _source: ['script_field_None']
 body: {'script_fields': {'script_field_None': {'script': {'source': "(((doc['AvgTicketPrice'].value * 2) + doc['DistanceKilometers'].value) - doc['FlightDelayMin'].value)"}}}}
 post_processing: []

>>> pd_df = ed.eland_to_pandas(df)
>>> pd_df.head()
   AvgTicketPrice  Cancelled           Carrier  ...        OriginWeather dayOfWeek           timestamp
0      841.265642      False   Kibana Airlines  ...                Sunny         0 2018-01-01 00:00:00
1      882.982662      False  Logstash Airways  ...                Clear         0 2018-01-01 18:27:00
2      190.636904      False  Logstash Airways  ...                 Rain         0 2018-01-01 17:11:14
3      181.694216       True   Kibana Airlines  ...  Thunder & Lightning         0 2018-01-01 10:33:28
4      730.041778      False   Kibana Airlines  ...        Damaging Wind         0 2018-01-01 05:13:00

[5 rows x 27 columns]
```

See [docs](https://eland.readthedocs.io/en/latest) and [demo_notebook.ipynb](https://eland.readthedocs.io/en/latest/examples/demo_notebook.html) for more examples.

## Where to get it
The source code is currently hosted on GitHub at:
https://github.com/elastic/eland

Binary installers for the latest released version are available at the [Python
package index](https://pypi.org/project/eland).

```sh
pip install eland
```
"""

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    maintainer=about['__maintainer__'],
    maintainer_email=about['__maintainer_email__'],
    license='Apache 2.0',
    classifiers=CLASSIFIERS,
    keywords='elastic eland pandas python',
    packages=find_packages(include=["eland", "eland.*"]),
    install_requires=[
        'elasticsearch>=7.0.5',
        'pandas==0.25.3',
        'matplotlib'
    ]
)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import eland as ed
from eland.tests import ES_TEST_CLIENT, FLIGHTS_INDEX_NAME
from eland.tests.common import TestData


class TestRename(TestData):

    def test_single_rename(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights_column_series = self.pd_flights().columns.to_series()

        assert pd_flights_column_series.index.to_list() == ed_field_mappings.display_names

        renames = {'DestWeather': 'renamed_DestWeather'}

        # inplace rename
        ed_field_mappings.rename(renames)

        assert pd_flights_column_series.rename(renames).index.to_list() == ed_field_mappings.display_names

        get_renames = ed_field_mappings.get_renames()

        assert renames == get_renames

    def test_non_exists_rename(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights_column_series = self.pd_flights().columns.to_series()

        assert pd_flights_column_series.index.to_list() == ed_field_mappings.display_names

        renames = {'unknown': 'renamed_unknown'}

        # inplace rename - in this case it has no effect
        ed_field_mappings.rename(renames)

        assert pd_flights_column_series.index.to_list() == ed_field_mappings.display_names

        get_renames = ed_field_mappings.get_renames()

        assert not get_renames

    def test_exists_and_non_exists_rename(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights_column_series = self.pd_flights().columns.to_series()

        assert pd_flights_column_series.index.to_list() == ed_field_mappings.display_names

        renames = {'unknown': 'renamed_unknown', 'DestWeather': 'renamed_DestWeather', 'unknown2': 'renamed_unknown2',
                   'Carrier': 'renamed_Carrier'}

        # inplace rename - only real names get renamed
        ed_field_mappings.rename(renames)

        assert pd_flights_column_series.rename(renames).index.to_list() == ed_field_mappings.display_names

        get_renames = ed_field_mappings.get_renames()

        assert {'Carrier': 'renamed_Carrier', 'DestWeather': 'renamed_DestWeather'} == get_renames

    def test_multi_rename(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights_column_series = self.pd_flights().columns.to_series()

        assert pd_flights_column_series.index.to_list() == ed_field_mappings.display_names

        renames = {'DestWeather': 'renamed_DestWeather', 'renamed_DestWeather': 'renamed_renamed_DestWeather'}

        # inplace rename - only first rename gets renamed
        ed_field_mappings.rename(renames)

        assert pd_flights_column_series.rename(renames).index.to_list() == ed_field_mappings.display_names

        get_renames = ed_field_mappings.get_renames()

        assert {'DestWeather': 'renamed_DestWeather'} == get_renames



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import eland as ed
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_INDEX_NAME
from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_series_equal


class TestSeriesHeadTail(TestData):

    def test_head_tail(self):
        pd_s = self.pd_flights()['Carrier']
        ed_s = ed.Series(ES_TEST_CLIENT, FLIGHTS_INDEX_NAME, 'Carrier')

        pd_s_head = pd_s.head(10)
        ed_s_head = ed_s.head(10)

        assert_pandas_eland_series_equal(pd_s_head, ed_s_head)

        pd_s_tail = pd_s.tail(10)
        ed_s_tail = ed_s.tail(10)

        assert_pandas_eland_series_equal(pd_s_tail, ed_s_tail)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.util.testing import assert_series_equal

from eland.tests.common import TestData


class TestDataFrameNUnique(TestData):

    def test_flights_nunique(self):
        # Note pandas.nunique fails for dict columns (e.g. DestLocation)
        columns = ['AvgTicketPrice', 'Cancelled', 'Carrier', 'Dest', 'DestAirportID', 'DestCityName']
        pd_flights = self.pd_flights()[columns]
        ed_flights = self.ed_flights()[columns]

        pd_nunique = pd_flights.nunique()
        ed_nunique = ed_flights.nunique()

        # TODO - ES is approximate counts so these aren't equal...
        # E[left]: [13059, 2, 4, 156, 156, 143]
        # E[right]: [13132, 2, 4, 156, 156, 143]
        # assert_series_equal(pd_nunique, ed_nunique)

    def test_ecommerce_nunique(self):
        columns = ['customer_first_name', 'customer_last_name', 'day_of_week_i']
        pd_ecommerce = self.pd_ecommerce()[columns]
        ed_ecommerce = self.ed_ecommerce()[columns]

        pd_nunique = pd_ecommerce.nunique()
        ed_nunique = ed_ecommerce.nunique()

        assert_series_equal(pd_nunique, ed_nunique)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

class Index:
    """
    The index for an eland.DataFrame.

    TODO - This currently has very different behaviour than pandas.Index

    Currently, the index is a field that exists in every document in an Elasticsearch index.
    For slicing and sorting operations it must be a docvalues field. By default _id is used,
    which can't be used for range queries and is inefficient for sorting:

    https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-id-field.html
    (The value of the _id field is also accessible in aggregations or for sorting,
    but doing so is discouraged as it requires to load a lot of data in memory.
    In case sorting or aggregating on the _id field is required, it is advised to duplicate
    the content of the _id field in another field that has doc_values enabled.)
    """
    ID_INDEX_FIELD = '_id'
    ID_SORT_FIELD = '_doc'  # if index field is _id, sort by _doc

    def __init__(self, query_compiler, index_field=None):
        self.index_field = index_field

        self._query_compiler = query_compiler

    @property
    def sort_field(self):
        if self._index_field == self.ID_INDEX_FIELD:
            return self.ID_SORT_FIELD
        return self._index_field

    @property
    def is_source_field(self):
        return self._is_source_field

    @property
    def index_field(self):
        return self._index_field

    @index_field.setter
    def index_field(self, index_field):
        if index_field is None or index_field == Index.ID_INDEX_FIELD:
            self._index_field = Index.ID_INDEX_FIELD
            self._is_source_field = False
        else:
            self._index_field = index_field
            self._is_source_field = True

    def __len__(self):
        return self._query_compiler._index_count()

    # Make iterable
    def __next__(self):
        # TODO resolve this hack to make this 'iterable'
        raise StopIteration()

    def __iter__(self):
        return self

    def info_es(self, buf):
        buf.write("Index:\n")
        buf.write(" index_field: {0}\n".format(self.index_field))
        buf.write(" is_source_field: {0}\n".format(self.is_source_field))



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
"""
Public plotting API

Based from https://github.com/pandas-dev/pandas/blob/v0.25.3/pandas/plotting/__init__.py
but only supporting a subset of plotting methods (for now).
"""
from eland.plotting._matplotlib.hist import (
    hist_frame,
    hist_series,
)

__all___ = [
    "hist_frame",
    "hist_series",
]



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import pandas as pd
import pytest

from eland.compat import PY36
from eland.dataframe import DEFAULT_NUM_ROWS_DISPLAYED
from eland.tests.common import TestData, assert_pandas_eland_series_equal


class TestDataFrameRepr(TestData):

    @classmethod
    def setup_class(cls):
        # conftest.py changes this default - restore to original setting
        pd.set_option('display.max_rows', 60)

    # Override these methods as:
    # {'lat': '-33.94609833', 'lon': '151.177002'} order is not consistent in python 3.5 (dict's not ordered)
    # remove from test for now
    def ed_flights(self):
        if not PY36:
            return super().ed_flights().drop(columns=['OriginLocation', 'DestLocation'])
        return super().ed_flights()

    def pd_flights(self):
        if not PY36:
            return super().pd_flights().drop(columns=['OriginLocation', 'DestLocation'])
        return super().pd_flights()

    """
    to_string
    """

    def test_simple_lat_lon(self):
        """
        Note on nested object order - this can change when
        note this could be a bug in ES...
        PUT my_index/doc/1
        {
          "location": {
            "lat": "50.033333",
            "lon": "8.570556"
          }
        }

        GET my_index/_search

        "_source": {
          "location": {
            "lat": "50.033333",
            "lon": "8.570556"
          }
        }

        GET my_index/_search
        {
          "_source": "location"
        }

        "_source": {
          "location": {
            "lon": "8.570556",
            "lat": "50.033333"
          }
        }

        Hence we store the pandas df source json as 'lon', 'lat'
        """
        if PY36:
            pd_dest_location = self.pd_flights()['DestLocation'].head(1)
            ed_dest_location = self.ed_flights()['DestLocation'].head(1)

            assert_pandas_eland_series_equal(pd_dest_location, ed_dest_location)
        else:
            # NOOP
            assert True

    def test_num_rows_to_string(self):
        if PY36:
            # check setup works
            assert pd.get_option('display.max_rows') == 60

            # Test eland.DataFrame.to_string vs pandas.DataFrame.to_string
            # In pandas calling 'to_string' without max_rows set, will dump ALL rows

            # Test n-1, n, n+1 for edge cases
            self.num_rows_to_string(DEFAULT_NUM_ROWS_DISPLAYED - 1)
            self.num_rows_to_string(DEFAULT_NUM_ROWS_DISPLAYED)
            with pytest.warns(UserWarning):
                # UserWarning displayed by eland here (compare to pandas with max_rows set)
                self.num_rows_to_string(DEFAULT_NUM_ROWS_DISPLAYED + 1, None, DEFAULT_NUM_ROWS_DISPLAYED)

            # Test for where max_rows lt or gt num_rows
            self.num_rows_to_string(10, 5, 5)
            self.num_rows_to_string(100, 200, 200)
        else:
            # NOOP
            assert True

    def num_rows_to_string(self, rows, max_rows_eland=None, max_rows_pandas=None):
        ed_flights = self.ed_flights()[['DestLocation', 'OriginLocation']]
        pd_flights = self.pd_flights()[['DestLocation', 'OriginLocation']]

        ed_head = ed_flights.head(rows)
        pd_head = pd_flights.head(rows)

        ed_head_str = ed_head.to_string(max_rows=max_rows_eland)
        pd_head_str = pd_head.to_string(max_rows=max_rows_pandas)

        # print("\n", ed_head_str)
        # print("\n", pd_head_str)

        assert pd_head_str == ed_head_str

    def test_empty_dataframe_string(self):
        ed_ecom = self.ed_ecommerce()
        pd_ecom = self.pd_ecommerce()

        ed_ecom_s = ed_ecom[ed_ecom['currency'] == 'USD'].to_string()
        pd_ecom_s = pd_ecom[pd_ecom['currency'] == 'USD'].to_string()

        assert ed_ecom_s == pd_ecom_s

    """
    repr
    """

    def test_num_rows_repr(self):
        self.num_rows_repr(pd.get_option('display.max_rows') - 1, pd.get_option('display.max_rows') - 1)
        self.num_rows_repr(pd.get_option('display.max_rows'), pd.get_option('display.max_rows'))
        self.num_rows_repr(pd.get_option('display.max_rows') + 1, pd.get_option('display.min_rows'))

    def num_rows_repr(self, rows, num_rows_printed):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head = ed_flights.head(rows)
        pd_head = pd_flights.head(rows)

        ed_head_str = repr(ed_head)
        pd_head_str = repr(pd_head)

        if num_rows_printed < rows:
            # add 1 for ellipsis
            num_rows_printed = num_rows_printed + 1

        # number of rows is num_rows_printed + 3 (header, summary)
        assert (num_rows_printed + 3) == len(ed_head_str.splitlines())

        assert pd_head_str == ed_head_str

    def test_empty_dataframe_repr(self):
        ed_ecom = self.ed_ecommerce()
        pd_ecom = self.pd_ecommerce()

        ed_ecom_r = repr(ed_ecom[ed_ecom['currency'] == 'USD'])
        pd_ecom_r = repr(pd_ecom[pd_ecom['currency'] == 'USD'])

        assert ed_ecom_r == pd_ecom_r

    """
    to_html
    """

    def test_num_rows_to_html(self):
        # check setup works
        assert pd.get_option('display.max_rows') == 60

        # Test eland.DataFrame.to_string vs pandas.DataFrame.to_string
        # In pandas calling 'to_string' without max_rows set, will dump ALL rows

        # Test n-1, n, n+1 for edge cases
        self.num_rows_to_html(DEFAULT_NUM_ROWS_DISPLAYED - 1)
        self.num_rows_to_html(DEFAULT_NUM_ROWS_DISPLAYED)
        with pytest.warns(UserWarning):
            # UserWarning displayed by eland here
            self.num_rows_to_html(DEFAULT_NUM_ROWS_DISPLAYED + 1, None, DEFAULT_NUM_ROWS_DISPLAYED)

        # Test for where max_rows lt or gt num_rows
        self.num_rows_to_html(10, 5, 5)
        self.num_rows_to_html(100, 200, 200)

    def num_rows_to_html(self, rows, max_rows_eland=None, max_rows_pandas=None):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head = ed_flights.head(rows)
        pd_head = pd_flights.head(rows)

        ed_head_str = ed_head.to_html(max_rows=max_rows_eland)
        pd_head_str = pd_head.to_html(max_rows=max_rows_pandas)

        # print(ed_head_str)
        # print(pd_head_str)

        assert pd_head_str == ed_head_str

    def test_empty_dataframe_to_html(self):
        ed_ecom = self.ed_ecommerce()
        pd_ecom = self.pd_ecommerce()

        ed_ecom_h = ed_ecom[ed_ecom['currency'] == 'USD'].to_html()
        pd_ecom_h = pd_ecom[pd_ecom['currency'] == 'USD'].to_html()

        assert ed_ecom_h == pd_ecom_h

    """
    _repr_html_
    """

    def test_num_rows_repr_html(self):
        # check setup works
        assert pd.get_option('display.max_rows') == 60

        show_dimensions = pd.get_option('display.show_dimensions')

        # TODO - there is a bug in 'show_dimensions' as it gets added after the last </div>
        # For now test without this
        pd.set_option('display.show_dimensions', False)

        # Test eland.DataFrame.to_string vs pandas.DataFrame.to_string
        # In pandas calling 'to_string' without max_rows set, will dump ALL rows

        # Test n-1, n, n+1 for edge cases
        self.num_rows_repr_html(pd.get_option('display.max_rows') - 1)
        self.num_rows_repr_html(pd.get_option('display.max_rows'))
        self.num_rows_repr_html(pd.get_option('display.max_rows') + 1, pd.get_option('display.max_rows'))

        # Restore default
        pd.set_option('display.show_dimensions', show_dimensions)

    def num_rows_repr_html(self, rows, max_rows=None):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_head = ed_flights.head(rows)
        pd_head = pd_flights.head(rows)

        ed_head_str = ed_head._repr_html_()
        pd_head_str = pd_head._repr_html_()

        # print(ed_head_str)
        # print(pd_head_str)

        # Currently pandas display bold_rows=True with >=PY36 and bold_rows=False with 3.5
        # TODO - fix this test for 3.5
        if PY36:
            assert pd_head_str == ed_head_str

    def test_empty_dataframe_repr_html(self):
        # TODO - there is a bug in 'show_dimensions' as it gets added after the last </div>
        # For now test without this
        show_dimensions = pd.get_option('display.show_dimensions')
        pd.set_option('display.show_dimensions', False)

        ed_ecom = self.ed_ecommerce()
        pd_ecom = self.pd_ecommerce()

        ed_ecom_rh = ed_ecom[ed_ecom['currency'] == 'USD']._repr_html_()
        pd_ecom_rh = pd_ecom[pd_ecom['currency'] == 'USD']._repr_html_()

        # Restore default
        pd.set_option('display.show_dimensions', show_dimensions)

        assert ed_ecom_rh == pd_ecom_rh



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import csv

import pandas as pd
from pandas.io.parsers import _c_parser_defaults

from eland import Client, DEFAULT_CHUNK_SIZE
from eland import DataFrame
from eland import FieldMappings


def read_es(es_client, es_index_pattern):
    """
    Utility method to create an eland.Dataframe from an Elasticsearch index_pattern.
    (Similar to pandas.read_csv, but source data is an Elasticsearch index rather than
    a csv file)

    Parameters
    ----------
    es_client: Elasticsearch client argument(s)
        - elasticsearch-py parameters or
        - elasticsearch-py instance or
        - eland.Client instance
    es_index_pattern: str
        Elasticsearch index pattern

    Returns
    -------
    eland.DataFrame

    See Also
    --------
    eland.pandas_to_eland: Create an eland.Dataframe from pandas.DataFrame
    eland.eland_to_pandas: Create a pandas.Dataframe from eland.DataFrame
    """
    return DataFrame(client=es_client, index_pattern=es_index_pattern)


def pandas_to_eland(pd_df,
                    es_client,
                    es_dest_index,
                    es_if_exists='fail',
                    es_refresh=False,
                    es_dropna=False,
                    es_geo_points=None,
                    chunksize=None):
    """
    Append a pandas DataFrame to an Elasticsearch index.
    Mainly used in testing.
    Modifies the elasticsearch destination index

    Parameters
    ----------
    es_client: Elasticsearch client argument(s)
        - elasticsearch-py parameters or
        - elasticsearch-py instance or
        - eland.Client instance
    es_dest_index: str
        Name of Elasticsearch index to be appended to
    es_if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the index already exists.

        - fail: Raise a ValueError.
        - replace: Delete the index before inserting new values.
        - append: Insert new values to the existing index. Create if does not exist.
    es_refresh: bool, default 'False'
        Refresh es_dest_index after bulk index
    es_dropna: bool, default 'False'
        * True: Remove missing values (see pandas.Series.dropna)
        * False: Include missing values - may cause bulk to fail
    es_geo_points: list, default None
        List of columns to map to geo_point data type
    chunksize: int, default None
        number of pandas.DataFrame rows to read before bulk index into Elasticsearch

    Returns
    -------
    eland.Dataframe
        eland.DataFrame referencing data in destination_index

    Examples
    --------

    >>> pd_df = pd.DataFrame(data={'A': 3.141,
    ...                            'B': 1,
    ...                            'C': 'foo',
    ...                            'D': pd.Timestamp('20190102'),
    ...                            'E': [1.0, 2.0, 3.0],
    ...                            'F': False,
    ...                            'G': [1, 2, 3]},
    ...                      index=['0', '1', '2'])
    >>> type(pd_df)
    <class 'pandas.core.frame.DataFrame'>
    >>> pd_df
           A  B  ...      F  G
    0  3.141  1  ...  False  1
    1  3.141  1  ...  False  2
    2  3.141  1  ...  False  3
    <BLANKLINE>
    [3 rows x 7 columns]
    >>> pd_df.dtypes
    A           float64
    B             int64
    C            object
    D    datetime64[ns]
    E           float64
    F              bool
    G             int64
    dtype: object

    Convert `pandas.DataFrame` to `eland.DataFrame` - this creates an Elasticsearch index called `pandas_to_eland`.
    Overwrite existing Elasticsearch index if it exists `if_exists="replace"`, and sync index so it is
    readable on return `refresh=True`


    >>> ed_df = ed.pandas_to_eland(pd_df,
    ...                            'localhost',
    ...                            'pandas_to_eland',
    ...                            es_if_exists="replace",
    ...                            es_refresh=True)
    >>> type(ed_df)
    <class 'eland.dataframe.DataFrame'>
    >>> ed_df
           A  B  ...      F  G
    0  3.141  1  ...  False  1
    1  3.141  1  ...  False  2
    2  3.141  1  ...  False  3
    <BLANKLINE>
    [3 rows x 7 columns]
    >>> ed_df.dtypes
    A           float64
    B             int64
    C            object
    D    datetime64[ns]
    E           float64
    F              bool
    G             int64
    dtype: object

    See Also
    --------
    eland.read_es: Create an eland.Dataframe from an Elasticsearch index
    eland.eland_to_pandas: Create a pandas.Dataframe from eland.DataFrame
    """
    if chunksize is None:
        chunksize = DEFAULT_CHUNK_SIZE

    client = Client(es_client)

    mapping = FieldMappings._generate_es_mappings(pd_df, es_geo_points)

    # If table exists, check if_exists parameter
    if client.index_exists(index=es_dest_index):
        if es_if_exists == "fail":
            raise ValueError(
                "Could not create the index [{0}] because it "
                "already exists. "
                "Change the if_exists parameter to "
                "'append' or 'replace' data.".format(es_dest_index)
            )
        elif es_if_exists == "replace":
            client.index_delete(index=es_dest_index)
            client.index_create(index=es_dest_index, body=mapping)
        # elif if_exists == "append":
        # TODO validate mapping are compatible
    else:
        client.index_create(index=es_dest_index, body=mapping)

    # Now add data
    actions = []
    n = 0
    for row in pd_df.iterrows():
        # Use index as _id
        id = row[0]

        if es_dropna:
            values = row[1].dropna().to_dict()
        else:
            values = row[1].to_dict()

        # Use integer as id field for repeatable results
        action = {'_index': es_dest_index, '_source': values, '_id': str(id)}

        actions.append(action)

        n = n + 1

        if n % chunksize == 0:
            client.bulk(actions, refresh=es_refresh)
            actions = []

    client.bulk(actions, refresh=es_refresh)

    ed_df = DataFrame(client, es_dest_index)

    return ed_df


def eland_to_pandas(ed_df, show_progress=False):
    """
    Convert an eland.Dataframe to a pandas.DataFrame

    **Note: this loads the entire Elasticsearch index into in core pandas.DataFrame structures. For large
    indices this can create significant load on the Elasticsearch cluster and require signficant memory**

    Parameters
    ----------
    ed_df: eland.DataFrame
        The source eland.Dataframe referencing the Elasticsearch index
    show_progress: bool
        Output progress of option to stdout? By default False.

    Returns
    -------
    pandas.Dataframe
        pandas.DataFrame contains all rows and columns in eland.DataFrame

    Examples
    --------
    >>> ed_df = ed.DataFrame('localhost', 'flights').head()
    >>> type(ed_df)
    <class 'eland.dataframe.DataFrame'>
    >>> ed_df
       AvgTicketPrice  Cancelled  ... dayOfWeek           timestamp
    0      841.265642      False  ...         0 2018-01-01 00:00:00
    1      882.982662      False  ...         0 2018-01-01 18:27:00
    2      190.636904      False  ...         0 2018-01-01 17:11:14
    3      181.694216       True  ...         0 2018-01-01 10:33:28
    4      730.041778      False  ...         0 2018-01-01 05:13:00
    <BLANKLINE>
    [5 rows x 27 columns]

    Convert `eland.DataFrame` to `pandas.DataFrame` (Note: this loads entire Elasticsearch index into core memory)

    >>> pd_df = ed.eland_to_pandas(ed_df)
    >>> type(pd_df)
    <class 'pandas.core.frame.DataFrame'>
    >>> pd_df
       AvgTicketPrice  Cancelled  ... dayOfWeek           timestamp
    0      841.265642      False  ...         0 2018-01-01 00:00:00
    1      882.982662      False  ...         0 2018-01-01 18:27:00
    2      190.636904      False  ...         0 2018-01-01 17:11:14
    3      181.694216       True  ...         0 2018-01-01 10:33:28
    4      730.041778      False  ...         0 2018-01-01 05:13:00
    <BLANKLINE>
    [5 rows x 27 columns]

    Convert `eland.DataFrame` to `pandas.DataFrame` and show progress every 10000 rows

    >>> pd_df = ed.eland_to_pandas(ed.DataFrame('localhost', 'flights'), show_progress=True) # doctest: +SKIP
    2020-01-29 12:43:36.572395: read 10000 rows
    2020-01-29 12:43:37.309031: read 13059 rows

    See Also
    --------
    eland.read_es: Create an eland.Dataframe from an Elasticsearch index
    eland.pandas_to_eland: Create an eland.Dataframe from pandas.DataFrame
    """
    return ed_df._to_pandas(show_progress=show_progress)


def read_csv(filepath_or_buffer,
             es_client,
             es_dest_index,
             es_if_exists='fail',
             es_refresh=False,
             es_dropna=False,
             es_geo_points=None,
             sep=",",
             delimiter=None,
             # Column and Index Locations and Names
             header="infer",
             names=None,
             index_col=None,
             usecols=None,
             squeeze=False,
             prefix=None,
             mangle_dupe_cols=True,
             # General Parsing Configuration
             dtype=None,
             engine=None,
             converters=None,
             true_values=None,
             false_values=None,
             skipinitialspace=False,
             skiprows=None,
             skipfooter=0,
             nrows=None,
             # Iteration
             # iterator=False,
             chunksize=None,
             # NA and Missing Data Handling
             na_values=None,
             keep_default_na=True,
             na_filter=True,
             verbose=False,
             skip_blank_lines=True,
             # Datetime Handling
             parse_dates=False,
             infer_datetime_format=False,
             keep_date_col=False,
             date_parser=None,
             dayfirst=False,
             cache_dates=True,
             # Quoting, Compression, and File Format
             compression="infer",
             thousands=None,
             decimal=b".",
             lineterminator=None,
             quotechar='"',
             quoting=csv.QUOTE_MINIMAL,
             doublequote=True,
             escapechar=None,
             comment=None,
             encoding=None,
             dialect=None,
             # Error Handling
             error_bad_lines=True,
             warn_bad_lines=True,
             # Internal
             delim_whitespace=False,
             low_memory=_c_parser_defaults["low_memory"],
             memory_map=False,
             float_precision=None):
    """
    Read a comma-separated values (csv) file into eland.DataFrame (i.e. an Elasticsearch index).

    **Modifies an Elasticsearch index**

    **Note pandas iteration options not supported**

    Parameters
    ----------
    es_client: Elasticsearch client argument(s)
        - elasticsearch-py parameters or
        - elasticsearch-py instance or
        - eland.Client instance
    es_dest_index: str
        Name of Elasticsearch index to be appended to
    es_if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the index already exists.

        - fail: Raise a ValueError.
        - replace: Delete the index before inserting new values.
        - append: Insert new values to the existing index. Create if does not exist.
    es_dropna: bool, default 'False'
        * True: Remove missing values (see pandas.Series.dropna)
        * False: Include missing values - may cause bulk to fail
    es_geo_points: list, default None
        List of columns to map to geo_point data type
    chunksize
        number of csv rows to read before bulk index into Elasticsearch

    Other Parameters
    ----------------
    Parameters derived from :pandas_api_docs:`pandas.read_csv`.

    See Also
    --------
    :pandas_api_docs:`pandas.read_csv`

    Notes
    -----
    iterator not supported

    Examples
    --------

    See if 'churn' index exists in Elasticsearch

    >>> from elasticsearch import Elasticsearch # doctest: +SKIP
    >>> es = Elasticsearch() # doctest: +SKIP
    >>> es.indices.exists(index="churn") # doctest: +SKIP
    False

    Read 'churn.csv' and use first column as _id (and eland.DataFrame index)
    ::

        # churn.csv
        ,state,account length,area code,phone number,international plan,voice mail plan,number vmail messages,total day minutes,total day calls,total day charge,total eve minutes,total eve calls,total eve charge,total night minutes,total night calls,total night charge,total intl minutes,total intl calls,total intl charge,customer service calls,churn
        0,KS,128,415,382-4657,no,yes,25,265.1,110,45.07,197.4,99,16.78,244.7,91,11.01,10.0,3,2.7,1,0
        1,OH,107,415,371-7191,no,yes,26,161.6,123,27.47,195.5,103,16.62,254.4,103,11.45,13.7,3,3.7,1,0
        ...

    >>>  ed.read_csv("churn.csv",
    ...             es_client='localhost',
    ...             es_dest_index='churn',
    ...             es_refresh=True,
    ...             index_col=0) # doctest: +SKIP
              account length  area code  churn  customer service calls  ... total night calls  total night charge total night minutes voice mail plan
    0                128        415      0                       1  ...                91               11.01               244.7             yes
    1                107        415      0                       1  ...               103               11.45               254.4             yes
    2                137        415      0                       0  ...               104                7.32               162.6              no
    3                 84        408      0                       2  ...                89                8.86               196.9              no
    4                 75        415      0                       3  ...               121                8.41               186.9              no
    ...              ...        ...    ...                     ...  ...               ...                 ...                 ...             ...
    3328             192        415      0                       2  ...                83               12.56               279.1             yes
    3329              68        415      0                       3  ...               123                8.61               191.3              no
    3330              28        510      0                       2  ...                91                8.64               191.9              no
    3331             184        510      0                       2  ...               137                6.26               139.2              no
    3332              74        415      0                       0  ...                77               10.86               241.4             yes
    <BLANKLINE>
    [3333 rows x 21 columns]

    Validate data now exists in 'churn' index:

    >>> es.search(index="churn", size=1) # doctest: +SKIP
    {'took': 1, 'timed_out': False, '_shards': {'total': 1, 'successful': 1, 'skipped': 0, 'failed': 0}, 'hits': {'total': {'value': 3333, 'relation': 'eq'}, 'max_score': 1.0, 'hits': [{'_index': 'churn', '_id': '0', '_score': 1.0, '_source': {'state': 'KS', 'account length': 128, 'area code': 415, 'phone number': '382-4657', 'international plan': 'no', 'voice mail plan': 'yes', 'number vmail messages': 25, 'total day minutes': 265.1, 'total day calls': 110, 'total day charge': 45.07, 'total eve minutes': 197.4, 'total eve calls': 99, 'total eve charge': 16.78, 'total night minutes': 244.7, 'total night calls': 91, 'total night charge': 11.01, 'total intl minutes': 10.0, 'total intl calls': 3, 'total intl charge': 2.7, 'customer service calls': 1, 'churn': 0}}]}}

    TODO - currently the eland.DataFrame may not retain the order of the data in the csv.
    """
    kwds = dict()

    kwds.update(
        sep=sep,
        delimiter=delimiter,
        engine=engine,
        dialect=dialect,
        compression=compression,
        # engine_specified=engine_specified,
        doublequote=doublequote,
        escapechar=escapechar,
        quotechar=quotechar,
        quoting=quoting,
        skipinitialspace=skipinitialspace,
        lineterminator=lineterminator,
        header=header,
        index_col=index_col,
        names=names,
        prefix=prefix,
        skiprows=skiprows,
        skipfooter=skipfooter,
        na_values=na_values,
        true_values=true_values,
        false_values=false_values,
        keep_default_na=keep_default_na,
        thousands=thousands,
        comment=comment,
        decimal=decimal,
        parse_dates=parse_dates,
        keep_date_col=keep_date_col,
        dayfirst=dayfirst,
        date_parser=date_parser,
        cache_dates=cache_dates,
        nrows=nrows,
        # iterator=iterator,
        chunksize=chunksize,
        converters=converters,
        dtype=dtype,
        usecols=usecols,
        verbose=verbose,
        encoding=encoding,
        squeeze=squeeze,
        memory_map=memory_map,
        float_precision=float_precision,
        na_filter=na_filter,
        delim_whitespace=delim_whitespace,
        warn_bad_lines=warn_bad_lines,
        error_bad_lines=error_bad_lines,
        low_memory=low_memory,
        mangle_dupe_cols=mangle_dupe_cols,
        infer_datetime_format=infer_datetime_format,
        skip_blank_lines=skip_blank_lines,
    )

    if chunksize is None:
        kwds.update(chunksize=DEFAULT_CHUNK_SIZE)

    client = Client(es_client)

    # read csv in chunks to pandas DataFrame and dump to eland DataFrame (and Elasticsearch)
    reader = pd.read_csv(filepath_or_buffer, **kwds)

    first_write = True
    for chunk in reader:
        if first_write:
            pandas_to_eland(chunk, client, es_dest_index, es_if_exists=es_if_exists, chunksize=chunksize,
                            es_refresh=es_refresh, es_dropna=es_dropna, es_geo_points=es_geo_points)
            first_write = False
        else:
            pandas_to_eland(chunk, client, es_dest_index, es_if_exists='append', chunksize=chunksize,
                            es_refresh=es_refresh, es_dropna=es_dropna, es_geo_points=es_geo_points)

    # Now create an eland.DataFrame that references the new index
    ed_df = DataFrame(client, es_dest_index)

    return ed_df



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.util.testing import assert_series_equal

from eland.tests.common import TestData


class TestDataFrameCount(TestData):

    def test_ecommerce_count(self):
        pd_ecommerce = self.pd_ecommerce()
        ed_ecommerce = self.ed_ecommerce()

        pd_count = pd_ecommerce.count()
        ed_count = ed_ecommerce.count()

        print(pd_count)
        print(ed_count)

        assert_series_equal(pd_count, ed_count)
